/* mrftopbm - convert mrf to pbm
 * public domain by RJM
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>


int bitbox,bitsleft;


void bit_init()
{
bitbox=0; bitsleft=0;
}


int bit_input(FILE *in)
{
if(bitsleft==0)
  {
  bitbox=fgetc(in);
  bitsleft=8;
  }

bitsleft--;
return((bitbox&(1<<bitsleft))?1:0);
}


void do_square(FILE *in,unsigned char *image,int ox,int oy,int w,int size)
{
int x,y,c;

/* is it all black or white? */

if(size==1 || bit_input(in))
  {
  /* yes, next bit says which. */
  c=bit_input(in);
  for(y=0;y<size;y++)
    for(x=0;x<size;x++)
      image[(oy+y)*w+ox+x]=c;
  }
else
  {
  /* not all one colour, so recurse. */
  size>>=1;
  do_square(in,image,ox,oy,w,size);
  do_square(in,image,ox+size,oy,w,size);
  do_square(in,image,ox,oy+size,w,size);
  do_square(in,image,ox+size,oy+size,w,size);
  }
}



int main(int argc,char *argv[])
{
static unsigned char buf[128];
FILE *in,*out;
unsigned char *image;
int expand_all=0;
int w,h,w8,w64,h64;
int x,y,c;
int i,mask;

in=stdin;
out=stdout;

if(argc>=2 && strcmp(argv[1],"-a")==0)
  {
  expand_all=1;
  argc--,argv++;
  }

if(argc>3 || (argc==2 && strcmp(argv[1],"-h")==0))
  {
  fprintf(stderr,"usage: mrftopbm <in.mrf >out.pbm\n");
  fprintf(stderr,"   or  mrftopbm  in.mrf >out.pbm\n");
  fprintf(stderr,"   or  mrftopbm  in.mrf  out.pbm\n");
  exit(1);
  }

if(argc==2 || argc==3)
  {
  if((in=fopen(argv[1],"rb"))==NULL)
    fprintf(stderr,"mrftopbm: couldn't open `%s'.\n",argv[1]),exit(1);
  if(argc==3)
    if((out=fopen(argv[2],"wb"))==NULL)
      fprintf(stderr,"mrftopbm: couldn't open `%s'.\n",argv[2]),exit(1);
  }

fread(buf,1,13,in);

if(strncmp(buf,"MRF1",4)!=0)
  fprintf(stderr,"mrftopbm: not an mrf file.\n"),exit(1);

if(buf[12]!=0)
  fprintf(stderr,"mrftopbm: can't handle this file subtype.\n"),exit(1);

w=((buf[4]<<24)|(buf[5]<<16)|(buf[6]<<8)|buf[7]);
h=((buf[8]<<24)|(buf[9]<<16)|(buf[10]<<8)|buf[11]);

/* w64 is units-of-64-bits width, h64 same for height */
w64=(w+63)/64;
h64=(h+63)/64;
if(expand_all) w=w64*64,h=h64*64;
w8=(w+7)/8;		/* bytes per PBM line */

if((image=calloc(w64*h64*64*64,1))==NULL)
  {
  fprintf(stderr,"mrftopbm: out of memory.\n");
  if(in!=stdin) fclose(in);
  if(out!=stdout) fclose(out);
  exit(1);
  }

fprintf(out,"P4\n%d %d\n",w,h);

/* now recursively input squares. */

/* init bit input */
bit_init();

for(y=0;y<h64;y++)
  for(x=0;x<w64;x++)
    do_square(in,image,x*64,y*64,w64*64,64);

if(in!=stdin) fclose(in);

/* write image */
for(y=0;y<h;y++)
  for(x=0;x<w8;x++)
    {
    c=255;
    for(i=0,mask=128;i<8;i++,mask>>=1)
      c^=mask*image[y*w64*64+x*8+i];
    fputc(c,out);
    }

free(image);

if(out!=stdout) fclose(out);

exit(0);
}
